#define INCL_DOS
#define INCL_DOSDEVIOCTL
#define INCL_DOSERRORS
#define INCL_NOPMAPI
#include <os2.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#if HAVE_ERRNO_H
#include <errno.h>
#endif

#include "et/com_err.h"
#include "ext2fs/ext2_err.h"
#include "io.h"

#define FLAGS_DASD   1		/* 1 : OPEN_FLAGS_DASD open         0 : normal open */
#define FLAGS_LOCKED 2		/* 1 : IOCtl DSK_LOCKDRIVE called   0 : not called  */
#define FLAGS_FORMAT 4          /* 1 : IOCtl DSK_BEGINFORMAT called 0 : not called  */

struct unix_private_data {
	int	magic;
	int	dev;
	int	flags;
	char	*buf;
	int	buf_block_nr;
#ifdef OS2
	int os2_flags;
#endif
};

static char *fs_name = "ext2";
int os2_open(const char *name, int flags, struct unix_private_data *data) {
    APIRET rc;
    HFILE  f;
    int namelen;
    int open_flags;
    int rw;
    ULONG action;
    ULONG parmio;
    ULONG dataio;
    ULONG pio;
    ULONG dio;

    namelen = strlen(name);
    printf("\tOS2 : os2_open %s %d\n", name, namelen);

    data->os2_flags = 0;
    /*
     * Case where "name" is a drive letter (block device name) : OS/2 has special requirements.
     */
    if ((namelen == 2) && (name[1] == ':') && (((name[0] >= 'A') && (name[0] <= 'Z')) || ((name[0] >= 'a') && (name[0] <= 'z')))) {
        rw = flags & IO_FLAG_RW ? OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYWRITE : OPEN_ACCESS_READONLY | OPEN_SHARE_DENYNONE;

        printf("\tOS2 : opening %s in DASD mode\n", name);
        if ((rc = DosOpen(
                      name, 
                      &f, 
                      &action, 
                      0, 
                      0, 
                      OPEN_ACTION_OPEN_IF_EXISTS, 
                      OPEN_FLAGS_DASD | OPEN_FLAGS_NO_CACHE | OPEN_FLAGS_WRITE_THROUGH  | rw, 
                      NULL
                     )) == NO_ERROR) {
            data->os2_flags = FLAGS_DASD;
            printf("\tOS2 : DosOpen OK\n");
            /*
             * In case the drive is mounted RW, we must lock the drive.
             */
            if (flags & IO_FLAG_RW) {
                parmio = 0;
                dataio = 0;
                pio    = 1;
                dio    = 1;
                if ((rc = DosDevIOCtl(
                                      f,
                                      IOCTL_DISK,
                                      DSK_LOCKDRIVE,
                                      &parmio, 1, &pio,
                                      &dataio, 1, &dio
                                     )) != NO_ERROR) {
                    os2_err(rc);
                    DosClose(f);
                    f = -1;
                    errno = EACCES;                   
                } else {
                    data->os2_flags |= FLAGS_LOCKED;

                    parmio = (ULONG)fs_name;
                    pio    = sizeof(char *);
                    dio    = 1;
                    dataio = 0;
                    if ((rc = DosDevIOCtl(
                                          f,
                                          IOCTL_DISK,
                                          DSK_BEGINFORMAT,
                                          &parmio, pio, &pio,
                                          &dataio, dio, &dio
                                         )) != NO_ERROR) {
                        os2_err(rc);
                        DosClose(f);
                        f = -1;
                        errno = EACCES;
                    } else {
                        data->os2_flags |= FLAGS_FORMAT;
                    }
                }
            }
        } else {
            os2_err(rc);                                              
            f = -1;
            errno = EACCES;
        }
        return f;
    }

    /*
     * Case where "name" is a regular file name or a character device name
     */
    return open(name, ((flags & IO_FLAG_RW) ? O_RDWR : O_RDONLY) | O_BINARY);

}

int os2_close(int fd, struct unix_private_data *data) {
    APIRET rc;
    ULONG parmio;
    ULONG dataio;
    ULONG pio;
    ULONG dio;
    
    if ((data->os2_flags & FLAGS_DASD) && (data->os2_flags & FLAGS_FORMAT)) {
        printf("\tOS2 : Redetermining media\n");
        parmio = 0;
        dataio = 0;
        pio = 1;
        dio = 1;
        rc = DosDevIOCtl(
                         fd,
                         IOCTL_DISK,
                         DSK_REDETERMINEMEDIA,
                         &parmio, pio, &pio,
                         &dataio, dio, &dio
                        );
         if (rc != NO_ERROR) {
             os2_err(rc);
         }
    }

    if ((data->os2_flags & FLAGS_DASD) && (data->os2_flags & FLAGS_LOCKED)) {
        printf("\tOS2 : Unlocking drive\n");
        parmio = 0;
        dataio = 0;
        pio = 1;
        dio = 1;
        rc = DosDevIOCtl(
                         fd,
                         IOCTL_DISK,
                         DSK_UNLOCKDRIVE,
                         &parmio, pio, &pio,
                         &dataio, dio, &dio
                        );
         if (rc != NO_ERROR) {
             os2_err(rc);
         }
    }

    printf("\tOS2 : Closing drive\n");
    return close(fd);
}
